﻿namespace artiso.TFSEventWorkflows.TFSEventWorkflowsServerPlugin
{
    #region *** using namepaces ***
    using System;
    using System.Activities;
    using System.Activities.XamlIntegration;
    using System.Collections.Generic;
    using System.Configuration;
    using System.IO;
    using System.Linq;
    using System.Reflection;
    using System.Xml;
    using System.Xml.Serialization;

    // TFS server object model
    using Microsoft.TeamFoundation.Framework.Common;
    using Microsoft.TeamFoundation.Framework.Server;
    using Microsoft.TeamFoundation.Server.Core;
    using Microsoft.TeamFoundation.VersionControl.Server;
    using Microsoft.TeamFoundation.WorkItemTracking.Server;

    using artiso.TFSEventWorkflows.LoggingLib;

    using log4net.Config;

    // TFS client object model - explicit using to avoid conflicts with the server object model
    using TfsTeamProjectCollection = Microsoft.TeamFoundation.Client.TfsTeamProjectCollection;
    using ITeamFoundationRegistry = Microsoft.TeamFoundation.Framework.Client.ITeamFoundationRegistry;
    using VersionControlServer = Microsoft.TeamFoundation.VersionControl.Client.VersionControlServer;
    using WorkingFolder = Microsoft.TeamFoundation.VersionControl.Client.WorkingFolder;
    using VersionSpec = Microsoft.TeamFoundation.VersionControl.Client.VersionSpec;
    using DeletedState = Microsoft.TeamFoundation.VersionControl.Client.DeletedState;
    using ItemType = Microsoft.TeamFoundation.VersionControl.Client.ItemType;
    using WorkingFolderType = Microsoft.TeamFoundation.VersionControl.Client.WorkingFolderType;
    using RecursionType = Microsoft.TeamFoundation.VersionControl.Client.RecursionType;
    using Workspace = Microsoft.TeamFoundation.VersionControl.Client.Workspace;
    using WorkspaceNotFoundException = Microsoft.TeamFoundation.VersionControl.Client.WorkspaceNotFoundException;
    using CreateWorkspaceParameters = Microsoft.TeamFoundation.VersionControl.Client.CreateWorkspaceParameters;
    using WorkspaceLocation = Microsoft.TeamFoundation.VersionControl.Common.WorkspaceLocation;
    using WorkspaceOptions = Microsoft.TeamFoundation.VersionControl.Common.WorkspaceOptions;
    using GetStatus = Microsoft.TeamFoundation.VersionControl.Client.GetStatus;
    using TfsConfigurationServer = Microsoft.TeamFoundation.Client.TfsConfigurationServer;
    #endregion

    internal class WorkflowRunner
    {
        #region *** Constants and Fields ***

        private FileInfo executionPath;

        private List<TFSEvent> _tfsEvents = null;
        private List<TFSEvent> TfsEvents(TeamFoundationRequestContext requestContext) 
        {
          if (_tfsEvents == null)
          {
            this.LogInfo(string.Format("Starting Update Cache"));

            List<TFSEvent> _tfsEventsFile = GetTFSEventConfigFromFile();
            List<TFSEvent> _tfsEventsTfs = GetTFSEventConfigFromTfs(requestContext);

            List<TFSEvent> __tfsEvents = new List<TFSEvent>();
            __tfsEvents.AddRange(_tfsEventsFile);
            __tfsEvents.AddRange(_tfsEventsTfs);
            _tfsEvents = __tfsEvents;

            _tfsEvents.ForEach((TFSEvent tfsEvent) =>
            {
              LogEvent(tfsEvent.Disabled ? "Disabled" : "Watching", tfsEvent);
            });

            this.LogInfo(string.Format("Leave Update Cache"));
          }
          return _tfsEvents;
        } 

        private Assembly configurationDefiningAssembly;

        private int CacheChangeSetId = 0;

        #endregion

        #region *** Constructors and Destructors ***

        public WorkflowRunner()
        {
        }

        #endregion

        #region *** Public Properties ***

        /// <summary>
        /// Gets the execution path of the assembly.
        /// </summary>
        /// <value>The execution path.</value>
        public FileInfo ExecutionPath
        {
            get
            {
                if (this.executionPath == null)
                {
                    this.executionPath = new FileInfo(new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath);
                }
                return this.executionPath;
            }
        }

        #endregion

        #region *** event processing ***

        private bool InJobAgent = false;

        public EventNotificationStatus ProcessEvent(TeamFoundationRequestContext requestContext, object notificationEventArgs)
        {
          Type typeNotification = notificationEventArgs.GetType();
          this.LogInfo(string.Format("Detecting event [{0}] defined in [{1}] - [{2}] defined in [{3}]", 
            typeNotification.Name, System.IO.Path.GetFileName(typeNotification.Assembly.Location),
            typeNotification.FullName, typeNotification.Assembly.Location));

          InJobAgent = (requestContext.ServiceName == "Team Foundation JobAgent");

          string strCollection = requestContext.ServiceHost.Name;
          string strProject = null;

          if (typeNotification == typeof(TFSEventAsync))
          {
            // that's a TfsEventWorkflows internal notification send from Web Service to Job Agent only
            System.Diagnostics.Debug.Assert(InJobAgent);
            
            // deserialize the job parameters tp avoid long term reloading of this values
            var notification = notificationEventArgs as TFSEventAsync;
            notification.tfsEvent.EventType = notification.notificationEventArgs.GetType();
            notification.tfsEvent.Async = false;
            _tfsEvents = new List<TFSEvent>();
            _tfsEvents.Add(notification.tfsEvent);
            CachedWorkflowFolders = notification.CachedWorkflowFolders;
            typeNotification = notification.notificationEventArgs.GetType();
            notificationEventArgs = notification.notificationEventArgs;
          }

          if (typeNotification == typeof(Microsoft.TeamFoundation.WorkItemTracking.Server.WorkItemChangedEvent))
          {
            // workitem - is assigned to one project
            var notification = notificationEventArgs as Microsoft.TeamFoundation.WorkItemTracking.Server.WorkItemChangedEvent;
            strProject = notification.PortfolioProject;
          }
          else if (typeNotification == typeof(Microsoft.TeamFoundation.VersionControl.Server.CheckinNotification))
          {
            // checkin - may contain items releated to different projects - don't set project
            var notification = notificationEventArgs as Microsoft.TeamFoundation.VersionControl.Server.CheckinNotification;

            // check whether the change set contains any cached TfsEvetWorkflows files
            bool UpdateNeeded = true;
            if (CachedWorkflowFolders != null)
            {
              UpdateNeeded = false;
              var cache = CachedWorkflowFolders.Where((CachedWorkflowFolder cachedWorkflowFolder) =>
                {
                  return cachedWorkflowFolder.PathServerColletion == strCollection;
                }).Select((CachedWorkflowFolder cachedWorkflowFolder) =>
                  {
                    return cachedWorkflowFolder.PathServer;
                  });

              UpdateNeeded = notification.SubmittedItems.Any((string SubmittedItem) =>
              {
                return cache.Any((string PathServer) => { return SubmittedItem.StartsWith(PathServer); });
              });
            }

            if (UpdateNeeded)
            {
              foreach (var SubmittedItem in notification.SubmittedItems)
                this.LogInfo(string.Format("checkin in workspace detected - {0}", SubmittedItem));

              // Update Cache
              _tfsEvents = null;
            }
          }
          else if (typeNotification == typeof(Microsoft.TeamFoundation.Build.Server.BuildQualityChangedNotificationEvent))
          {
            // build - is assigned to one project
            var notification = notificationEventArgs as Microsoft.TeamFoundation.Build.Server.BuildQualityChangedNotificationEvent;
            strProject = notification.Build.TeamProject;
          }
          else if (typeNotification == typeof(Microsoft.TeamFoundation.Build.Server.BuildCompletionNotificationEvent))
          {
            // build - is assigned to one project
            var notification = notificationEventArgs as Microsoft.TeamFoundation.Build.Server.BuildCompletionNotificationEvent;
            strProject = notification.Build.TeamProject;
          }
          else if (typeNotification == typeof(Microsoft.TeamFoundation.Build.Server.BuildDefinitionChangedEvent))
          {
            // build - is assigned to one project
            var notification = notificationEventArgs as Microsoft.TeamFoundation.Build.Server.BuildDefinitionChangedEvent;
            strProject = notification.TeamProject;
          }
          else if (typeNotification == typeof(Microsoft.TeamFoundation.TestManagement.Server.TestSuiteChangedNotification))
          {
            // testsuite - is assigned to one project
            var notification = notificationEventArgs as Microsoft.TeamFoundation.TestManagement.Server.TestSuiteChangedNotification;
            strProject = notification.ProjectName;
          }

          List<TFSEvent> TfsEvents = this.TfsEvents(requestContext);

          var TfsEventsMatching = TfsEvents.Where((TFSEvent tfsEvent) =>
          {
            return 
              (!tfsEvent.Disabled && notificationEventArgs.GetType() == tfsEvent.EventType) &&
              (string.IsNullOrEmpty(tfsEvent.Collection) || tfsEvent.Collection == strCollection) &&
              (string.IsNullOrEmpty(tfsEvent.Project) || string.IsNullOrEmpty(strProject) || tfsEvent.Project == strProject);
          });
          if(TfsEventsMatching.Count() != 0)
            this.LogInfo(string.Format("Handling event on collection [{0}] project [{1}] - {2} workflows match", strCollection, string.IsNullOrEmpty(strProject) ? "<not provided>" : strProject, TfsEventsMatching.Count())); 


          foreach (TFSEvent tfsEvent in TfsEvents)
          {
            if (!tfsEvent.Disabled && notificationEventArgs.GetType() == tfsEvent.EventType)
            {
              if (string.IsNullOrEmpty(tfsEvent.Collection) || tfsEvent.Collection == strCollection)
              {
                if (string.IsNullOrEmpty(tfsEvent.Project) || string.IsNullOrEmpty(strProject) || tfsEvent.Project == strProject)
                {
                  if (!tfsEvent.Async)
                  {
                    try
                    {
                      // to enable assemlies probing in cached files got from version control
                      AppDomain.CurrentDomain.AssemblyResolve += BinariesResolveEventHandler;

                      // get the name of the workflow file
                      var workflow = this.GetWorkflow(tfsEvent.WorkflowFileName);

                      // initialize the workflow arguments
                      // TODO: provide more comfort - check the arguments defined in the workflow and provide the information requested by the workflow -> makes it easier for the workflow
                      var workflowParameters = new Dictionary<string, object>
                            {
                                { "TFSEvent", notificationEventArgs },
                                { "TeamFoundationRequestContext", requestContext }
                            };

                      // invoke the workflow
                      LogEvent("Starting", tfsEvent);
                      WorkflowInvoker.Invoke(workflow, workflowParameters);
                    }
                    catch (Exception e)
                    {
                      this.LogError(string.Format("Workflow error:"), e);
                    }
                    finally
                    {
                      // to disable assemlies probing in cached files got from version control
                      AppDomain.CurrentDomain.AssemblyResolve -= BinariesResolveEventHandler;

                      LogEvent("Leaving", tfsEvent);
                    }
                  }
                  else
                  {
                    try
                    {
                      // provide the job some parameters to avoid long term reloading of this values
                      TFSEventAsync async = new TFSEventAsync()
                      {
                        notificationEventArgs = notificationEventArgs,
                        tfsEvent = tfsEvent,
                        CachedWorkflowFolders = CachedWorkflowFolders,
                      };

                      XmlNode xmlData = SerializeXml(async);

                      // handle the notification by queueing the information we need for a job
                      LogEvent("Queuing", tfsEvent);
                      var jobService = requestContext.GetService<TeamFoundationJobService>();
                      var jobGuid = jobService.QueueOneTimeJob(
                          requestContext,
                          "TFSEventWorkflow Job",
                          "artiso.TFSEventWorkflows.TFSEventWorkflowsServerPlugin.WorkflowRunnerJob",
                          xmlData,
                          false);
                      int result = jobService.QueueJobsNow(requestContext, new List<Guid>() { jobGuid }, true);
                    }
                    catch (Exception e)
                    {
                      this.LogError(string.Format("Queuing error:"), e);
                    }
                  }
                }
              }
            }
          }

          if (TfsEventsMatching.Count() != 0)
            this.LogInfo("");
          return EventNotificationStatus.ActionPermitted;
        }

        /// <summary>
        /// Subscribeds the types.
        /// </summary>
        /// <returns></returns>
        public Type[] SubscribedTypes()
        {
          return new Type[]
          {
            // workitem tracking
            typeof(Microsoft.TeamFoundation.WorkItemTracking.Server.WorkItemChangedEvent),
            // version control
            typeof(Microsoft.TeamFoundation.VersionControl.Server.CheckinNotification),
            // build
            typeof(Microsoft.TeamFoundation.Build.Server.BuildQualityChangedNotificationEvent),
            typeof(Microsoft.TeamFoundation.Build.Server.BuildCompletionNotificationEvent),
            typeof(Microsoft.TeamFoundation.Build.Server.BuildDefinitionChangedEvent),
            // testing
            typeof(Microsoft.TeamFoundation.TestManagement.Server.TestPlanChangedNotification),
            typeof(Microsoft.TeamFoundation.TestManagement.Server.TestSuiteChangedNotification),
            typeof(Microsoft.TeamFoundation.TestManagement.Server.TestConfigurationChangedNotification),
            typeof(Microsoft.TeamFoundation.TestManagement.Server.TestRunStartedNotification),
            typeof(Microsoft.TeamFoundation.TestManagement.Server.TestRunChangedNotification),
            typeof(Microsoft.TeamFoundation.TestManagement.Server.TestRunCompletedNotification),
          };
        }

        #endregion

        #region *** helpers ***

        private void LogEvent(string action, TFSEvent tfsEvent)
        {
          this.LogInfo(string.Format("{0} event [{1}] workflow file [{2}] defined in config [{3}] - workflow file [{4}] defined in config [{5}])",
            action,
            tfsEvent.Name,
            System.IO.Path.GetFileName(tfsEvent.WorkflowFileName), System.IO.Path.GetFileName(tfsEvent.ConfigFileName),
            tfsEvent.WorkflowFileName, tfsEvent.ConfigFileName
            ));
        }

        private XmlNode SerializeXml(TFSEventAsync notificationEventArgs)
        {
          if (null == notificationEventArgs)
          {
            return null;
          }

          Type typeNotificationEventArgs = notificationEventArgs.GetType();
          var xmlSerializer = new XmlSerializer(typeNotificationEventArgs, new Type[] { (notificationEventArgs as TFSEventAsync).notificationEventArgs.GetType() });
          var memoryStream = new MemoryStream();
          xmlSerializer.Serialize(memoryStream, notificationEventArgs);

          memoryStream.Position = 0;
          var xmlDocument = new XmlDocument();
          xmlDocument.Load(memoryStream);
          memoryStream.Close();

          XmlNode xmlNode = xmlDocument.ChildNodes[1];
          XmlAttribute xmlAttribute = xmlDocument.CreateAttribute("assemblyQualifiedName");
          xmlAttribute.Value = typeNotificationEventArgs.AssemblyQualifiedName + "|" + notificationEventArgs.notificationEventArgs.GetType().AssemblyQualifiedName;
          xmlNode.Attributes.Append(xmlAttribute);

          return xmlNode;
        }

        #endregion

        #region *** helpers loading configuation from tfs

        private List<CachedWorkflowFolder> CachedWorkflowFolders = null;

        /// <summary>
        /// Gets the TFS event config.
        /// </summary>
        private List<TFSEvent> GetTFSEventConfigFromTfs(TeamFoundationRequestContext requestContextCollection)
        {
          List<TFSEvent> _tfsEvents = new List<TFSEvent>();

          try
          {
            // services on the collection
            TeamFoundationRegistryService srvRegistryCollection = requestContextCollection.GetService<TeamFoundationRegistryService>();
            TeamFoundationLocationService srvLocation = requestContextCollection.GetService<TeamFoundationLocationService>();

            // services on the configuration
            TeamFoundationRequestContext requestContextConfiguration = requestContextCollection.ServiceHost.ApplicationServiceHost.CreateServicingContext();
            TeamFoundationRegistryService srvRegistryConfiguration = requestContextConfiguration.GetService<TeamFoundationRegistryService>();

            // setup a registry notification on collection specific reg keys to refresh cache on registry change
            srvRegistryCollection.RegisterNotification(requestContextCollection, new RegistrySettingsChangedCallback(RegistryChanged_Collection), new string[] 
            { 
              // TODO: verify how to do this for ALL collection, not only for the collection corresponding to the current request
              TfsRegistryPath.WorkflowsCollectionPath,
            });

            // setup a registry notification on configuration specific reg keys to refresh cache on registry change
            srvRegistryConfiguration.RegisterNotification(requestContextConfiguration, new RegistrySettingsChangedCallback(RegistryChanged_Configuration), new string[] 
            { 
              TfsRegistryPath.WorkflowActivitiesGlobalCollection,
              TfsRegistryPath.WorkflowActivitiesGlobalPath,
              TfsRegistryPath.WorkflowsGlobalCollection,
              TfsRegistryPath.WorkflowsGlobalPath,
              TfsRegistryPath.WorkflowsProjectReleativePath,
              TfsRegistryPath.WorkflowsConfigFileNameWebService,
            });

            // unregister notification for change in registry key for sync with other app tiers to avoid recursion
            srvRegistryConfiguration.UnregisterNotification(requestContextConfiguration, RegistryChanged_ConfigurationCacheChangeSet);

            // patch the server uri to get exactly this app tier
            Uri uriRequestContextCollection = new Uri(srvLocation.GetPublicAccessMapping(requestContextCollection).AccessPoint);
            UriBuilder uriBuilderRequestContextCollection = new UriBuilder(uriRequestContextCollection.Scheme, System.Environment.MachineName, uriRequestContextCollection.Port, uriRequestContextCollection.PathAndQuery);
            uriRequestContextCollection = uriBuilderRequestContextCollection.Uri;

            // get access to tfs and collection
            Uri uriTfsServer = new Uri(uriRequestContextCollection.AbsoluteUri);
            Uri uriTfsCollection = new Uri(uriRequestContextCollection.AbsoluteUri + "/" + requestContextCollection.ServiceHost.Name);

            // read the configuration registry values
            RegistryValues regValues = ReadTfsEventWorkflowSettings(requestContextConfiguration, srvRegistryConfiguration);
            this.LogInfo(string.Format("{0}: {1}", "WorkflowActivitiesGlobalCollection", regValues.WorkflowActivitiesGlobalCollection));
            this.LogInfo(string.Format("{0}: {1}", "WorkflowActivitiesGlobalPath", regValues.WorkflowActivitiesGlobalPath));
            this.LogInfo(string.Format("{0}: {1}", "WorkflowsGlobalCollection", regValues.WorkflowsGlobalCollection));
            this.LogInfo(string.Format("{0}: {1}", "WorkflowsGlobalPath", regValues.WorkflowsGlobalPath));
            this.LogInfo(string.Format("{0}: {1}", "WorkflowsProjectRelativePath", regValues.WorkflowsProjectRelativePath));
            this.LogInfo(string.Format("{0}: {1}", "WorkflowsConfigFileName", regValues.WorkflowsConfigFileName));

            // local base path for version control workspace mapping
            string pathLocalWorkflowData = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData) + @"\TfsEventWorkflows";

            string strWorkspaceName = "TfsEventWorkflows-On-" + System.Environment.MachineName;
            string strWorkspaceOwner = Environment.UserName;
            string strWorkspaceComputer = Environment.MachineName;

            List<CachedWorkflowFolder> workflowFolders = new List<CachedWorkflowFolder>();

            VersionSpec vcVersionSpec = VersionSpec.Latest;
            DeletedState vcDeleteState = DeletedState.NonDeleted;
            ItemType vcItemType = ItemType.Folder;
            WorkingFolderType vcWorkFolderType = WorkingFolderType.Map;
            RecursionType vcRecursionType = RecursionType.Full;

            int idChangeSetMax = 0;

            TeamFoundationCatalogService srvCatalogConfiguration = requestContextConfiguration.GetService<TeamFoundationCatalogService>();

            // get all app tiers
            CatalogNode nodeRootInfrastructure = srvCatalogConfiguration.QueryRootNode(requestContextConfiguration, CatalogTree.Infrastructure);
            var nodesMachine = nodeRootInfrastructure.QueryChildren(requestContextConfiguration,
              new Guid[] { CatalogResourceTypes.Machine }, true, CatalogQueryOptions.None);

            foreach (var nodeMachine in nodesMachine)
            {
              this.LogInfo(string.Format("machine: {0}", nodeMachine.Resource.DisplayName));
              foreach(var prop in  nodeMachine.Resource.Properties)
              {
                this.LogInfo(string.Format("  {0}: {1}", prop.Key, prop.Value));
              }
            }

            // get all collections
            CatalogNode nodeRootOrganizational = srvCatalogConfiguration.QueryRootNode(requestContextConfiguration, CatalogTree.Organizational);
            var nodesCollection = nodeRootOrganizational.QueryChildren(requestContextConfiguration,
              new Guid[] { CatalogResourceTypes.ProjectCollection }, true, CatalogQueryOptions.None);

            foreach (var nodeCollection in nodesCollection)
            {
              string strCollection = nodeCollection.Resource.DisplayName;
              Guid quidCollection = new Guid(nodeCollection.Resource.Properties["InstanceId"]);
            }

            TfsConfigurationServer tfsConfigurationServer = new TfsConfigurationServer(uriTfsServer);
            var collectionNodes = tfsConfigurationServer.CatalogNode.QueryChildren(
                  new[] { CatalogResourceTypes.ProjectCollection }, false, CatalogQueryOptions.None);
            foreach (var collectionNode in collectionNodes)
            {
              // use client object model to access other collections
              string strCollection = collectionNode.Resource.DisplayName;
              Guid quidCollection = new Guid(collectionNode.Resource.Properties["InstanceId"]);
              TfsTeamProjectCollection tfsTeamProjectCollection = tfsConfigurationServer.GetTeamProjectCollection(quidCollection);
              ITeamFoundationRegistry tfsRegistry = tfsTeamProjectCollection.GetService<ITeamFoundationRegistry>();
              VersionControlServer tfsVersionControl = tfsTeamProjectCollection.GetService<VersionControlServer>();

              List<WorkingFolder> newWorkingFolders = new List<WorkingFolder>();

              // map workflow activities
              bool bWorkflowActivitiesGlobalConfigured = (!string.IsNullOrEmpty(regValues.WorkflowActivitiesGlobalCollection)) && !(string.IsNullOrEmpty(regValues.WorkflowActivitiesGlobalPath));
              if (bWorkflowActivitiesGlobalConfigured && (strCollection == regValues.WorkflowActivitiesGlobalCollection))
              {
                string pathServer = string.Format("$/{0}", regValues.WorkflowActivitiesGlobalPath);
                string pathLocal = string.Format(@"{0}\{1}\{2}", pathLocalWorkflowData, strCollection, regValues.WorkflowActivitiesGlobalPath.Replace(@"/", @"\"));

                if (tfsVersionControl.ServerItemExists(pathServer, vcVersionSpec, vcDeleteState, vcItemType))
                {
                  newWorkingFolders.Add(new WorkingFolder(pathServer, pathLocal, vcWorkFolderType, vcRecursionType));
                  workflowFolders.Add(new CachedWorkflowFolder(
                    CachedWorkflowFolderType.WorkflowActivitiesGlobal, null, null, pathServer, pathLocal, strCollection));
                }
              }
              // map global workflows
              bool bWorkflowsGlobalCollectionConfigured = (!string.IsNullOrEmpty(regValues.WorkflowsGlobalCollection)) && !(string.IsNullOrEmpty(regValues.WorkflowsGlobalPath));
              if (bWorkflowsGlobalCollectionConfigured && (strCollection == regValues.WorkflowsGlobalCollection))
              {
                string pathServer = string.Format("$/{0}", regValues.WorkflowsGlobalPath);
                string pathLocal = string.Format(@"{0}\{1}\{2}", pathLocalWorkflowData, strCollection, regValues.WorkflowsGlobalPath.Replace(@"/", @"\"));

                if (tfsVersionControl.ServerItemExists(pathServer, vcVersionSpec, vcDeleteState, vcItemType))
                {
                  newWorkingFolders.Add(new WorkingFolder(pathServer, pathLocal, vcWorkFolderType, vcRecursionType));
                  workflowFolders.Add(new CachedWorkflowFolder(
                    CachedWorkflowFolderType.WorkflowsGlobal, null, null, pathServer, pathLocal, strCollection));
                }
              }

              // map collection workflows
              string WorkflowsCollectionPath = tfsRegistry.GetValue(TfsRegistryPath.WorkflowsCollectionPath, "").Trim(new char[] { '/' });
              this.LogInfo(string.Format("collection: {0} - {1}: {2}", strCollection, "WorkflowsCollectionPath", WorkflowsCollectionPath));
              if (!string.IsNullOrEmpty(WorkflowsCollectionPath))
              {
                string pathServer = string.Format("$/{0}", WorkflowsCollectionPath);
                string pathLocal = string.Format(@"{0}\{1}\{2}", pathLocalWorkflowData, strCollection, WorkflowsCollectionPath.Replace(@"/", @"\"));

                if (tfsVersionControl.ServerItemExists(pathServer, vcVersionSpec, vcDeleteState, vcItemType))
                {
                  newWorkingFolders.Add(new WorkingFolder(pathServer, pathLocal, vcWorkFolderType, vcRecursionType));
                  workflowFolders.Add(new CachedWorkflowFolder(
                    CachedWorkflowFolderType.WorkflowsCollection, strCollection, null, pathServer, pathLocal, strCollection));
                }
              }

              // List the team projects in the collection
              bool bWorkflowsProjectRelativePathConfigured = (!string.IsNullOrEmpty(regValues.WorkflowsProjectRelativePath));
              if (bWorkflowsProjectRelativePathConfigured)
              {
                // get a catalog of team projects for the collection
                var projectNodes = collectionNode.QueryChildren(
                    new[] { CatalogResourceTypes.TeamProject }, false, CatalogQueryOptions.None);

                foreach (var projectNode in projectNodes)
                {
                  string strProject = projectNode.Resource.DisplayName;
                  string pathServer = string.Format("$/{0}/{1}", strProject, regValues.WorkflowsProjectRelativePath);
                  string pathLocal = string.Format(@"{0}\{1}\{2}\{3}", pathLocalWorkflowData, strCollection, strProject, regValues.WorkflowsProjectRelativePath.Replace(@"/", @"\"));

                  if (tfsVersionControl.ServerItemExists(pathServer, vcVersionSpec, vcDeleteState, vcItemType))
                  {
                    // map project workflows
                    newWorkingFolders.Add(new WorkingFolder(pathServer, pathLocal, vcWorkFolderType, vcRecursionType));
                    workflowFolders.Add(new CachedWorkflowFolder(
                      CachedWorkflowFolderType.WorkflowsProject, strCollection, strProject, pathServer, pathLocal, strCollection));
                  }
                }
              }

              // create/adjust workspace
              Workspace workspace = null;
              try
              {
                // get the workspace
                workspace = tfsVersionControl.GetWorkspace(strWorkspaceName, strWorkspaceOwner);
                // find changes
                var NewNotInOld = newWorkingFolders.Where(x1 => !workspace.Folders.Any(x2 => x2.ServerItem == x1.ServerItem));
                var OldNotInNew = workspace.Folders.Where(x1 => !newWorkingFolders.Any(x2 => x2.ServerItem == x1.ServerItem));
                // adjust folder mapping
                foreach (var map in OldNotInNew) workspace.DeleteMapping(map);
                foreach (var map in NewNotInOld) workspace.Map(map.ServerItem, map.LocalItem);
              }
              catch (WorkspaceNotFoundException)
              {
                // define workspace paraemters
                CreateWorkspaceParameters cwp = new CreateWorkspaceParameters(strWorkspaceName);
                cwp.Comment = "this workspace was created by TfsEventWorkflows. Please don't modify";
                cwp.Computer = strWorkspaceComputer;
                cwp.Folders = newWorkingFolders.ToArray();
                cwp.Location = WorkspaceLocation.Server;
                cwp.OwnerName = strWorkspaceOwner;
                cwp.WorkspaceName = strWorkspaceName;
                cwp.WorkspaceOptions = WorkspaceOptions.SetFileTimeToCheckin;
                // create new workspace
                workspace = tfsVersionControl.CreateWorkspace(cwp);
              }

              if (newWorkingFolders.Count > 0)
              {
                // TODO: verify this get - there is a problem if one of the custom activity assemblies is loaded and so blocked
                // get latest version
                var srvItems = newWorkingFolders.ConvertAll<string>(x => x.ServerItem).ToArray();
                GetStatus status = workspace.Get(srvItems, Microsoft.TeamFoundation.VersionControl.Client.VersionSpec.Latest, Microsoft.TeamFoundation.VersionControl.Client.RecursionType.Full, Microsoft.TeamFoundation.VersionControl.Client.GetOptions.Overwrite);

                newWorkingFolders.ForEach(folder =>
                {
                  var hist = tfsVersionControl.QueryHistory(folder.ServerItem, RecursionType.Full, 1);
                  if (hist.Count() > 0)
                  {
                    int idChangeset = hist.First().ChangesetId;
                    idChangeSetMax = Math.Max(idChangeSetMax, idChangeset);
                  }
                });
              }
            }

            // write idChangeSetMax to registry ==> this triggers a notification on all other app tiers
            CacheChangeSetId = idChangeSetMax;
            if (CacheChangeSetId > ReadTfsEventWorkflowCacheChangeSet(requestContextConfiguration, srvRegistryConfiguration))
              WriteTfsEventWorkflowCacheChangeSet(requestContextConfiguration, srvRegistryConfiguration, CacheChangeSetId);

            // register notification for change in registry key for sync with other app tiers
            srvRegistryConfiguration.RegisterNotification(requestContextCollection, new RegistrySettingsChangedCallback(RegistryChanged_ConfigurationCacheChangeSet), new string[] 
            { 
              TfsRegistryPath.CacheChangeSet,
            });

            // read all config files
            foreach (CachedWorkflowFolder workflowFolder in workflowFolders)
            {
              switch (workflowFolder.FolderType)
              {
                default:
                  break;
                case CachedWorkflowFolderType.WorkflowActivitiesGlobal:
                  break;
                case CachedWorkflowFolderType.WorkflowsGlobal:
                case CachedWorkflowFolderType.WorkflowsCollection:
                case CachedWorkflowFolderType.WorkflowsProject:
                  string strConfigFile = System.IO.Path.Combine(workflowFolder.PathLocal, regValues.WorkflowsConfigFileName);
                  if (System.IO.File.Exists(strConfigFile))
                  {
                    List<TFSEvent> __tfsEvents = GetTFSEventConfigFromFile(strConfigFile);
                    __tfsEvents.ForEach(x =>
                    {
                      x.WorkflowFileName = System.IO.Path.Combine(workflowFolder.PathLocal, System.IO.Path.GetFileName(x.WorkflowFileName));
                    });
                    _tfsEvents.AddRange(__tfsEvents);
                  }
                  break;
              }
            }

            CachedWorkflowFolders = workflowFolders;
          }
          catch (Exception ex)
          {
            this.LogError(string.Format("Loading error:"), ex);
          }

          return _tfsEvents;
        }

        #endregion

        #region *** helpers for loading configuration from file

        /// <summary>
        /// Gets the TFS event config.
        /// </summary>
        private List<TFSEvent> GetTFSEventConfigFromFile()
        {
          // load the config
          string strPluginFile = new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath;
          string strConfigFile = string.Format("{0}.config", strPluginFile);
          return GetTFSEventConfigFromFile(strConfigFile, strPluginFile);
        }

        /// <summary>
        /// Gets the TFS event config.
        /// </summary>
        private List<TFSEvent> GetTFSEventConfigFromFile(string strConfigFile)
        {
          // load the config
          string strPluginFile = new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath;
          return GetTFSEventConfigFromFile(strConfigFile, strPluginFile);
        }

        /// <summary>
        /// Gets the TFS event config.
        /// </summary>
        private List<TFSEvent> GetTFSEventConfigFromFile(string strConfigFile, string strPluginFile)
        {
          List<TFSEvent> _tfsEvents = new List<TFSEvent>();

          Configuration config = ConfigurationManager.OpenExeConfiguration(strPluginFile);

          TFSEventConfig tfsEventConfig;
          try
          {
              tfsEventConfig = (TFSEventConfig)config.GetSection("tfsEventConfig");
          }
          catch (ConfigurationErrorsException)
          {
              // if the config can not be loaded try to load it with reflection
              tfsEventConfig = GetCustomConfig<TFSEventConfig>(strPluginFile, strConfigFile, "tfsEventConfig");
          }

          tfsEventConfig = GetCustomConfig<TFSEventConfig>(strPluginFile, strConfigFile, "tfsEventConfig");
                
          _tfsEvents = (from TFSEventElement tfsEventCollection in tfsEventConfig.TFSEvents
                            select
                                new TFSEvent
                                    {
                                        Collection = tfsEventCollection.Collection,
                                        Project = tfsEventCollection.Project,
                                        FullTypeName = tfsEventCollection.FullTypeName,
                                        EventAssemblyName = tfsEventCollection.EventAssemblyName,
                                        WorkflowFileName = tfsEventCollection.WorkflowFileName,
                                        ConfigFileName = strConfigFile,
                                        Disabled = tfsEventCollection.Disabled,
                                        Async = tfsEventCollection.Async,
                                        EventType = null,
                                        Name = tfsEventCollection.Name
                                    }).ToList();

          // set the full type
          foreach (TFSEvent tfsEvent in _tfsEvents)
          {
              Assembly externalAssembly = Assembly.Load(tfsEvent.EventAssemblyName);
              tfsEvent.EventType = externalAssembly.GetType(tfsEvent.FullTypeName);

              if (!Path.IsPathRooted(tfsEvent.WorkflowFileName))
                tfsEvent.WorkflowFileName = this.ExecutionPath.Directory.FullName + "\\" + tfsEvent.WorkflowFileName;
          }
          
          return _tfsEvents;
        }

        #endregion

        #region *** config and assmebly loading helpers ***

        /// <summary>
        /// Gets the custom config.
        /// </summary>
        /// <typeparam name="TConfig">The type of the config.</typeparam>
        /// <param name="configDefiningAssemblyPath">The config defining assembly path.</param>
        /// <param name="configFilePath">The config file path.</param>
        /// <param name="sectionName">Name of the section.</param>
        /// <returns></returns>
        private TConfig GetCustomConfig<TConfig>(string configDefiningAssemblyPath, string configFilePath, string sectionName) where TConfig : ConfigurationSection
        {
            AppDomain.CurrentDomain.AssemblyResolve += this.ConfigResolveEventHandler;
            configurationDefiningAssembly = Assembly.LoadFrom(configDefiningAssemblyPath);
            var exeFileMap = new ExeConfigurationFileMap();
            exeFileMap.ExeConfigFilename = configFilePath;
            var customConfig = ConfigurationManager.OpenMappedExeConfiguration(exeFileMap, ConfigurationUserLevel.None);
            var returnConfig = customConfig.GetSection(sectionName) as TConfig;
            AppDomain.CurrentDomain.AssemblyResolve -= ConfigResolveEventHandler;
            return returnConfig;
        }

        private Assembly ConfigResolveEventHandler(object sender, ResolveEventArgs args)
        {
            return configurationDefiningAssembly;
        }

        private Assembly BinariesResolveEventHandler(object sender, ResolveEventArgs args)
        {
          CachedWorkflowFolder cachedWorkFolder = CachedWorkflowFolders.Find(x => x.FolderType == CachedWorkflowFolderType.WorkflowActivitiesGlobal);
          if(null == cachedWorkFolder) return null;

          string strWorkflowActivitiesGlobal = cachedWorkFolder.PathLocal;
          if(string.IsNullOrEmpty(strWorkflowActivitiesGlobal)) return null;
          
          // get all assemblies from WorkflowActivitiesGlobal (reflection only)
          Assembly[] cachedAssemblies = System.IO.Directory.EnumerateFiles(strWorkflowActivitiesGlobal).Select(x => 
          {
            try { return Assembly.ReflectionOnlyLoadFrom(x); }
            catch (Exception) { return null; }
          }).Where(x => x!=null).ToArray();

          // load the requested assembly
          Assembly cachedAssembly = cachedAssemblies.FirstOrDefault(x => { return x.GetName().Name == args.Name; });
          if (null != cachedAssembly)
          {
            this.LogInfo(string.Format("Loading Assembly [{0}]", cachedAssembly.Location));
            return Assembly.LoadFrom(cachedAssembly.Location);
          }

          return null;
        }

        /// <summary>
        /// Gets the workflow.
        /// </summary>
        /// <param name="workflowFileName">The workflow file.</param>
        /// <returns></returns>
        private Activity GetWorkflow(string workflowFileName)
        {
            // check whether file can be found as given in config
            if (!File.Exists(workflowFileName))
            {
                workflowFileName = this.ExecutionPath.Directory.FullName + "\\" + workflowFileName;
            }

            if (!File.Exists(workflowFileName))
            {
                throw new IOException(string.Format("Workflow file not found: {0}", workflowFileName));
            }

            Activity workflow = null;
            if (workflowFileName.EndsWith(".xaml"))
            {
                // for dynamic load from xaml
                workflow = ActivityXamlServices.Load(workflowFileName);
                //CheckWorkflowArguments(workflow as DynamicActivity);
                var d = workflow as DynamicActivity;
            }
            else if (workflowFileName.EndsWith(".dll"))
            {
                // for dynamic load from dll
                Assembly assm = Assembly.LoadFrom(workflowFileName);
                Type workflowType = null;
                foreach (Type typeInAssembly in assm.GetTypes())
                {
                    if (typeInAssembly.BaseType == typeof(Activity))
                    {
                        //CheckWorkflowArguments(typeInAssembly);
                        workflowType = typeInAssembly;
                        break;
                    }
                }
                workflow = (Activity)assm.CreateInstance(workflowType.ToString());
            }

            return workflow;
        }

        #endregion

        #region *** tfs registry helpers ***

        private struct RegistryValues
        {
          // workflow activities are stored in a central folder in the version control and distributed to each application tier
          public string WorkflowActivitiesGlobalCollection;
          public string WorkflowActivitiesGlobalPath;

          // global workflows are stored in a central folder in the version control and distributed to each application tier
          public string WorkflowsGlobalCollection;
          public string WorkflowsGlobalPath;

          // collection specific workflows are stored in a central folder in the version control and distributed to each application tier
          // public string WorkflowsCollectionPath; // stored in collection registry, not in configuration registry

          // collection and project specific workflows are stored in a central folder in the version control and distributed to each application tier
          public string WorkflowsProjectRelativePath;

          // name of the confif file - only workflows configured in this file are triggered
          public string WorkflowsConfigFileName;
        }

        // TFS registry paths
        class TfsRegistryPath
        {
          public const string Root = "/Configuration/Application/TfsEventWorkflows";
          public const string CacheChangeSet = Root + "/CacheChangeSet";
          public const string WorkflowActivitiesGlobalCollection = Root + "/WorkflowActivities/Global/Collection";
          public const string WorkflowActivitiesGlobalPath = Root + "/WorkflowActivities/Global/Path";
          public const string WorkflowsGlobalCollection = Root + "/Workflows/Global/Collection";
          public const string WorkflowsGlobalPath = Root + "/Workflows/Global/Path";
          public const string WorkflowsCollectionPath = Root + "/Workflows/Collection/Path";
          public const string WorkflowsProjectReleativePath = Root + "/Workflows/Project/PathRelative";
          public const string WorkflowsConfigFileNameWebService = Root + "/Workflows/ConfigFileName/WebService";
          public const string WorkflowsConfigFileNameJobAgent = Root + "/Workflows/ConfigFileName/JobAgent";
        }

        // read all configuation values
        private RegistryValues ReadTfsEventWorkflowSettings(TeamFoundationRequestContext requestContext, TeamFoundationRegistryService srvRegistry)
        {
          RegistryValues regValues = new RegistryValues()
          {
            WorkflowActivitiesGlobalCollection = srvRegistry.GetValue(requestContext, TfsRegistryPath.WorkflowActivitiesGlobalCollection, "").Trim(new char[] { '/' }),
            WorkflowActivitiesGlobalPath = srvRegistry.GetValue(requestContext, TfsRegistryPath.WorkflowActivitiesGlobalPath, "").Trim(new char[] { '/' }),

            WorkflowsGlobalCollection = srvRegistry.GetValue(requestContext, TfsRegistryPath.WorkflowsGlobalCollection, "").Trim(new char[] { '/' }),
            WorkflowsGlobalPath = srvRegistry.GetValue(requestContext, TfsRegistryPath.WorkflowsGlobalPath, "").Trim(new char[] { '/' }),
            WorkflowsProjectRelativePath = srvRegistry.GetValue(requestContext, TfsRegistryPath.WorkflowsProjectReleativePath, "").Trim(new char[] { '/' }),
            WorkflowsConfigFileName = srvRegistry.GetValue(requestContext, InJobAgent ? TfsRegistryPath.WorkflowsConfigFileNameJobAgent : TfsRegistryPath.WorkflowsConfigFileNameWebService, "").Trim(new char[] { '/' }),
          };

          return regValues;
        }

        // read cache change set id
        private int ReadTfsEventWorkflowCacheChangeSet(TeamFoundationRequestContext requestContext, TeamFoundationRegistryService srvRegistry)
        {
          return srvRegistry.GetValue<int>(requestContext, TfsRegistryPath.CacheChangeSet, -1);
        }

        // write cache change set id
        private void WriteTfsEventWorkflowCacheChangeSet(TeamFoundationRequestContext requestContext, TeamFoundationRegistryService srvRegistry, int idChangeSet)
        {
          srvRegistry.SetValue(requestContext, TfsRegistryPath.CacheChangeSet, idChangeSet);
        }

        private void RegistryChanged_Collection(TeamFoundationRequestContext requestContext, RegistryEntryCollection entries)
        {
          foreach (var entry in entries)
            this.LogInfo(string.Format("registry changed (collection) - {0} [{1}]", entry.Value, entry.Path));

          // someone has changed the collection registry under the TfsEventWorkflows hive
          _tfsEvents = null;
        }

        private void RegistryChanged_Configuration(TeamFoundationRequestContext requestContext, RegistryEntryCollection entries)
        {
          foreach (var entry in entries)
            this.LogInfo(string.Format("registry changed (configuration) - {0} [{1}]", entry.Value, entry.Path));

          // someone has changed the configuation registry under the TfsEventWorkflows hive
          _tfsEvents = null;
        }

        private void RegistryChanged_ConfigurationCacheChangeSet(TeamFoundationRequestContext requestContext, RegistryEntryCollection entries)
        {
          foreach (var entry in entries)
            this.LogInfo(string.Format("registry changed (configuration) - {0} [{1}]", entry.Value, entry.Path));

          // another application tier has detected a change of the cached files on the version control server
          _tfsEvents = null;
        }


        #endregion
    }
}